@@ -1,6 +1,6 @@
 package PLP;
 
-use v5.6;
+use 5.006;
 
 use PLP::Functions ();
 use PLP::Fields;
@@ -14,30 +14,172 @@ use Cwd ();
 
 use strict;
 
-our $VERSION = '3.16';
+our $VERSION = '3.17';
 
-# subs in this package:
-#  sendheaders                      Send headers
-#  source($path, $level, $linespec) Read and parse .plp files
-#  error($error, $type)             Handle errors
+# Subs in this package:
 #  _default_error($plain, $html)    Default error handler
-#  clean                            Reset variables
 #  cgi_init                         Initialization for CGI
-#  mod_perl_init($r)                Initialization for mod_perl
-#  start                            Start the initialized PLP script
+#  clean                            Reset variables
+#  error($error, $type)             Handle errors
 #  everything                       Do everything: CGI
 #  handler($r)                      Do everything: mod_perl
+#  mod_perl_init($r)                Initialization for mod_perl
+#  mod_perl_print		    Faster printing for mod_perl
+#  sendheaders                      Send headers
+#  source($path, $level, $linespec) Read and parse .plp files
+#  start                            Start the initialized PLP script
 
-# About the #S lines:
-# I wanted to implement Safe.pm so that scripts were run inside a
-# configurable compartment. This needed for XS modules to be pre-loaded,
-# hence the PLPsafe_* Apache directives. However, $safe->reval() lets
-# Apache segfault. End of fun. The lines are still here so that I can
-# s/^#S //m to re-implement them whenever this has been fixed.
+# The _init subs do the following:
+#  Set $PLP::code to the initial code
+#  Set $ENV{PLP_*} and makes PATH_INFO if needed
+#  Change the CWD
+
+# This gets referenced as the initial $PLP::ERROR
+sub _default_error {
+    my ($plain, $html) = @_; 
+    print qq{<table border=1 class="PLPerror"><tr><td>},
+	  qq{<span><b>Debug information:</b><BR>$html</td></tr></table>};
+}
+
+# CGI initializer: parses PATH_TRANSLATED
+sub cgi_init {
+
+    $PLP::print = 'print';
+    
+    my $path = $ENV{PATH_TRANSLATED};
+    $ENV{PLP_NAME} = $ENV{PATH_INFO};
+    my $path_info;
+    while (not -f $path) {
+        if (not $path =~ s/(\/+[^\/]*)$//) {
+	    print STDERR "PLP: Not found: $ENV{PATH_TRANSLATED} ($ENV{REQUEST_URI})\n";
+	    PLP::error(undef, 404);
+	    exit;
+	}
+	my $pi = $1;
+	$ENV{PLP_NAME} =~ s/\Q$pi\E$//;
+	$path_info = $pi . $path_info;
+    }
+    
+    if (not -r $path) {
+	print STDERR "PLP: Can't read: $ENV{PATH_TRANSLATED} ($ENV{REQUEST_URI})\n";
+	PLP::error(undef, 403);
+	exit;
+    }
+
+    delete @ENV{
+	qw(PATH_TRANSLATED SCRIPT_NAME SCRIPT_FILENAME PATH_INFO),
+        grep { /^REDIRECT_/ } keys %ENV
+    };
+
+    $ENV{PATH_INFO} = $path_info if defined $path_info;
+    $ENV{PLP_FILENAME} = $path;
+    my ($file, $dir) = File::Basename::fileparse($path);
+    chdir $dir;
+
+    $PLP::code = PLP::source($file, 0, undef, $path);
+}
+
+# This cleans up from previous requests, and sets the default $PLP::DEBUG
+sub clean {
+    @PLP::END = ();
+    $PLP::code = '';
+    $PLP::sentheaders = 0;
+    $PLP::DEBUG = 1;
+    $PLP::print = '';
+    $PLP::r = undef;
+    delete @ENV{ grep /^PLP_/, keys %ENV };
+}
+
+# Handles errors, uses subref $PLP::ERROR (default: \&_default_error)
+sub error {
+    my ($error, $type) = @_;
+    if (not defined $type or $type < 100) {
+	return undef unless $PLP::DEBUG & 1;
+	my $plain = $error;
+	(my $html = $plain) =~ s/([<&>])/'&#' . ord($1) . ';'/ge;
+	PLP::sendheaders() unless $PLP::sentheaders;
+	$PLP::ERROR->($plain, $html);
+    } else {
+	select STDOUT;
+	my ($short, $long) = @{
+	    +{
+		404 => [
+		    'Not Found',
+		    "The requested URL $ENV{REQUEST_URI} was not found " .
+		    "on this server."
+		],
+		403 => [
+		    'Forbidden',
+		    "You don't have permission to access $ENV{REQUEST_URI} " .
+		    "on this server."
+		],
+	    }->{$type}
+	};
+	print "Status: $type\nContent-Type: text/html\n\n",
+	      qq{<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">\n<html>},
+	      "<head>\n<title>$type $short</title>\n</head></body>\n<h1>$short",
+	      "</h1>\n$long<p>\n<hr>\n$ENV{SERVER_SIGNATURE}</body></html>";
+    }
+}
+
+# This is run by the CGI script. (#!perl \n use PLP; PLP::everything;)
+sub everything {
+    clean();
+    cgi_init();
+    start();
+}
+
+# This is the mod_perl handler.
+sub handler {
+    require Apache::Constants;
+    clean();
+    if (my $ret = mod_perl_init($_[0])) {
+	return $ret;
+    }
+    #S start($_[0]);
+    start();
+    no strict 'subs';
+    return Apache::Constants::OK();
+}
+
+# mod_perl initializer: returns 0 on success, Apache error code on failure
+sub mod_perl_init {
+    our $r = shift;
+
+    $PLP::print = 'PLP::mod_perl_print';
+    
+    $ENV{PLP_FILENAME} = my $filename = $r->filename;
+    
+    unless (-f $filename) {
+	return Apache::Constants::NOT_FOUND();
+    }
+    unless (-r _) {
+	return Apache::Constants::FORBIDDEN();
+    }
+    
+    $ENV{PLP_NAME} = $r->uri;
+
+    our $use_cache = $r->dir_config('PLPcache') !~ /^off$/i;
+#S  our $use_safe  = $r->dir_config('PLPsafe')  =~ /^on$/i;
+    my $path = $r->filename();
+    my ($file, $dir) = File::Basename::fileparse($path);
+    chdir $dir;
+
+    $PLP::code = PLP::source($file, 0, undef, $path);
+
+    return 0; # OK
+}
+
+# FAST printing under mod_perl
+sub mod_perl_print {
+    return if @_ == 1 and not length $_[0];
+    PLP::sendheaders() unless $PLP::sentheaders;
+    $PLP::r->print(@_);
+}
 
 # Sends the headers waiting in %PLP::Script::header
 sub sendheaders () {
-    our $sentheaders = 1;
+    $PLP::sentheaders ||= [ caller 1 ? (caller 1)[1, 2] : (caller)[1, 2] ];
     print STDOUT "Content-Type: text/plain\n\n" if $PLP::DEBUG & 2;
     print STDOUT map("$_: $PLP::Script::header{$_}\n", keys %PLP::Script::header), "\n";
 }
@@ -50,10 +192,13 @@ sub sendheaders () {
     # this function parses a PLP file and returns Perl code, ready to be eval'ed
     sub source {
 	my ($file, $level, $linespec, $path) = @_;
+	our $use_cache;
+
 	# $file is displayed, $path is used. $path is constructed from $file if
 	# not given.
-	$level = 0      if not defined $level;
-	$linespec = '1' if not defined $linespec;
+
+	$level = 0      unless defined $level;
+	$linespec = '1' unless defined $linespec;
 	
 	if ($level > 128) {
 	    %cached = ();
@@ -62,12 +207,13 @@ sub sendheaders () {
 		: qq{\n#line $linespec\ndie qq[Include recursion detected];};
 	}
 
-	our ($inA, $inB, $use_cache);
+	my $in_block = 0;   # 1 => "<:", 2 => "<:="
+	
 	$path ||= File::Spec->rel2abs($file);
 	
 	my $source_start = $level
-	    ? qq/\cQ;\n#line 1 "$file"\nprint q\cQ/
-	    : qq/\n#line 1 "$file"\nprint q\cQ/;
+	    ? qq/\cQ;\n#line 1 "$file"\n$PLP::print q\cQ/
+	    : qq/\n#line 1 "$file"\n$PLP::print q\cQ/;
 	
 	if ($use_cache and exists $cached{$path}) {
 	    BREAKOUT: {
@@ -105,7 +251,7 @@ sub sendheaders () {
 		    \G                  # Begin where left off
 		    ( \z                # End
 		    | <:=? | :>         # PLP tags     <:= ... :> <: ... :>
-		    | <\(.*?\)>         # Include tags <(...)>
+		    | <\([^)]*\)>         # Include tags <(...)>
 		    | <[^:(][^<:]*      # Normal text
 		    | :[^>][^<:]*       # Normal text
 		    | [^<:]*            # Normal text
@@ -113,25 +259,28 @@ sub sendheaders () {
 		/gxs;
 		next LINE unless length $1;
 		my $part = $1;
-		if ($part eq '<:=' and not $inA || $inB) {
-		    $inA = 1;
+		if ($part eq '<:=' and not $in_block) {
+		    $in_block = 2;
 		    $source .= "\cQ, ";
-		} elsif ($part eq '<:' and not $inA || $inB) {
-		    $inB = 1;
+		} elsif ($part eq '<:' and not $in_block) {
+		    $in_block = 1;
 		    $source .= "\cQ; ";
-		} elsif ($part eq ':>' and $inA) {
-		    $inA = 0;
-		    $source .= ", q\cQ";
-		} elsif ($part eq ':>' and $inB) {
-		    $inB = 0;
-		    $source .= "; print q\cQ";
-		} elsif ($part =~ /^<\((.*?)\)>\z/ and not $inA || $inB) {
-		    my $ipath = File::Spec->rel2abs($1, File::Basename::dirname($path));
+		} elsif ($part eq ':>' and $in_block) {
+		    $source .= (
+			  $in_block == 2
+			? ", q\cQ"               # 2
+			: "; $PLP::print q\cQ"   # 1
+		    );
+		    $in_block = 0;
+		} elsif ($part =~ /^<\((.*?)\)>\z/ and not $in_block) {
+		    my $ipath = File::Spec->rel2abs(
+			$1, File::Basename::dirname($path)
+		    );
 		    $source .= source($1, $level + 1, undef, $ipath) .
 			       qq/\cQ, \n#line $linenr "$file"\nq\cQ/;
 		    push @{ $cached{$path}[0] }, $ipath;
 		} else {
-		    $part =~ s/\\/\\\\/ if not $inA || $inB;
+		    $part =~ s/\\/\\\\/ unless $in_block;
 		    $source .= $part;
 		}
 	    }
@@ -148,141 +297,9 @@ sub sendheaders () {
     }
 }
 
-# Handles errors, uses subref $PLP::ERROR (default: \&_default_error)
-sub error {
-    my ($error, $type) = @_;
-    if (not defined $type or $type < 100) {
-	return undef unless $PLP::DEBUG & 1;
-	my $plain = $error;
-	(my $html = $plain) =~ s/([<&>])/'&#' . ord($1) . ';'/ge;
-	PLP::sendheaders unless $PLP::sentheaders;
-	$PLP::ERROR->($plain, $html);
-    } else {
-	select STDOUT;
-	my ($short, $long) = @{
-	    +{
-		404 => [
-		    'Not Found',
-		    "The requested URL $ENV{REQUEST_URI} was not found on this server."
-		],
-		403 => [
-		    'Forbidden',
-		    "You don't have permission to access $ENV{REQUEST_URI} on this server."
-		],
-	    }->{$type}
-	};
-	print "Status: $type\nContent-Type: text/html\n\n",
-	      qq{<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">\n},
-	      "<html><head>\n<title>--$type $short</title>\n</head></body>\n",
-	      "<h1>$short</h1>\n$long<p>\n<hr>\n$ENV{SERVER_SIGNATURE}</body></html>";
-    }
-}
-
-# This gets referenced as the initial $PLP::ERROR
-sub _default_error {
-    my ($plain, $html) = @_; 
-    print qq{<table border=1 class="PLPerror"><tr><td>},
-	  qq{<span><b>Debug information:</b><BR>$html</td></tr></table>};
-}
-
-# This cleans up from previous requests, and sets the default $PLP::DEBUG
-sub clean {
-    @PLP::END = ();
-    $PLP::code = '';
-    $PLP::sentheaders = 0;
-    $PLP::inA = 0;
-    $PLP::inB = 0;
-    $PLP::DEBUG = 1;
-    delete @ENV{ grep /^PLP_/, keys %ENV };
-}
-
-# The *_init subs do the following:
-#  o  Set $PLP::code to the initial code
-#  o  Set $ENV{PLP_*} and makes PATH_INFO if needed
-#  o  Change the CWD
-
-# CGI initializer: parses PATH_TRANSLATED
-sub cgi_init {
-    my $path = $ENV{PATH_TRANSLATED};
-    $ENV{PLP_NAME} = $ENV{PATH_INFO};
-    my $path_info;
-    while (not -f $path) {
-        if (not $path =~ s/(\/+[^\/]*)$//) {
-	    print STDERR "PLP: Not found: $ENV{PATH_TRANSLATED} ($ENV{REQUEST_URI})\n";
-	    PLP::error(undef, 404);
-	    exit;
-	}
-	my $pi = $1;
-	$ENV{PLP_NAME} =~ s/\Q$pi\E$//;
-	$path_info = $pi . $path_info;
-    }
-    
-    if (not -r $path) {
-	print STDERR "PLP: Can't read: $ENV{PATH_TRANSLATED} ($ENV{REQUEST_URI})\n";
-	PLP::error(undef, 403);
-	exit;
-    }
-
-    delete @ENV{
-	qw(PATH_TRANSLATED SCRIPT_NAME SCRIPT_FILENAME PATH_INFO),
-        grep { /^REDIRECT_/ } keys %ENV
-    };
-
-    $ENV{PATH_INFO} = $path_info if defined $path_info;
-    $ENV{PLP_FILENAME} = $path;
-    my ($file, $dir) = File::Basename::fileparse($path);
-    chdir $dir;
-
-    $PLP::code = PLP::source($file, 0, undef, $path);
-}
-
-# mod_perl initializer: returns 0 on success, Apache error code on failure
-sub mod_perl_init {
-    my $r = shift;
-    
-    $ENV{PLP_FILENAME} = my $filename = $r->filename;
-    
-    unless (-f $filename) {
-	return Apache::Constants::NOT_FOUND();
-    }
-    unless (-r _) {
-	return Apache::Constants::FORBIDDEN();
-    }
-    
-    $ENV{PLP_NAME} = $r->uri;
-
-    our $use_cache = $r->dir_config('PLPcache') !~ /^off$/i;
-#S  our $use_safe  = $r->dir_config('PLPsafe')  =~ /^on$/i;
-    my $path = $r->filename();
-    my ($file, $dir) = File::Basename::fileparse($path);
-    chdir $dir;
-
-    $PLP::code = PLP::source($file, 0, undef, $path);
-
-    return 0; # OK
-}
-
-#S # For PLPsafe scripts
-#S sub safe_eval {
-#S     my ($r, $code) = @_;
-#S     $r->send_http_header('text/plain');
-#S     require Safe;
-#S     unless ($PLP::safe) {
-#S 	$PLP::safe = Safe->new('PLP::Script');
-#S 	for ( map split, $r->dir_config->get('PLPsafe_module') ) {
-#S 	    $PLP::safe->share('*' . $_ . '::');
-#S 	    s!::!/!g;
-#S 	    require $_ . '.pm';
-#S 	}
-#S 	$PLP::safe->permit(Opcode::full_opset());
-#S 	$PLP::safe->deny(Opcode::opset(':dangerous'));
-#S     }
-#S     $PLP::safe->reval($code);
-#S }
 
 # Let the games begin! No lexicals may exist at this point.
 sub start {
-#S  my ($r) = @_;
     no strict;
     tie *PLPOUT, 'PLP::Tie::Print';
     select PLPOUT;
@@ -295,47 +312,20 @@ sub start {
 	*headers = \%header;
 	*cookies = \%cookie;
 	PLP::Functions->import();
+
 	# No lexicals may exist at this point.
 	
-#S 	if ($PLP::use_safe) {
-#S 	    PLP::safe_eval($r, $PLP::code);
-#S 	} else {
-	    eval qq{ package PLP::Script; $PLP::code; };
-#S 	}
+	eval qq{ package PLP::Script; $PLP::code; };
 	PLP::error($@, 1) if $@ and $@ !~ /\cS\cT\cO\cP/;
 
-#S 	if ($PLP::use_safe) {
-#S 	    PLP::safe_eval($r, '$_->() for reverse @PLP::END');
-#S 	} else {
-	    eval   { package PLP::Script; $_->() for reverse @PLP::END };
-#S 	}
+	eval   { package PLP::Script; $_->() for reverse @PLP::END };
 	PLP::error($@, 1) if $@ and $@ !~ /\cS\cT\cO\cP/;
     }
     PLP::sendheaders() unless $PLP::sentheaders;
     select STDOUT;
     undef *{"PLP::Script::$_"} for keys %PLP::Script::;
-#    Symbol::delete_package('PLP::Script');
-#    The above does not work. TODO - find out why not.
-}
-
-# This is run by the CGI script. (#!perl \n use PLP; PLP::everything;)
-sub everything {
-    clean();
-    cgi_init();
-    start();
-}
-
-# This is the mod_perl handler.
-sub handler {
-    require Apache::Constants;
-    clean();
-    if (my $ret = mod_perl_init($_[0])) {
-	return $ret;
-    }
-#S  start($_[0]);
-    start();
-    no strict 'subs';
-    return Apache::Constants::OK();
+    # Symbol::delete_package('PLP::Script');
+    # The above does not work. TODO - find out why not.
 }
 
 1;
@@ -545,3 +535,46 @@ L<PLP::Functions>, L<PLP::Fields>, L<PLP::FAQ>, L<PLP::HowTo>
 
 =cut
 
+### Garbage bin
+
+# About the #S lines:
+# I wanted to implement Safe.pm so that scripts were run inside a
+# configurable compartment. This needed for XS modules to be pre-loaded,
+# hence the PLPsafe_* Apache directives. However, $safe->reval() lets
+# Apache segfault. End of fun. The lines are still here so that I can
+# s/^#S //g to re-implement them whenever this has been fixed.
+
+#S # For PLPsafe scripts
+#S sub safe_eval {
+#S     my ($r, $code) = @_;
+#S     $r->send_http_header('text/plain');
+#S     require Safe;
+#S     unless ($PLP::safe) {
+#S 	$PLP::safe = Safe->new('PLP::Script');
+#S 	for ( map split, $r->dir_config->get('PLPsafe_module') ) {
+#S 	    $PLP::safe->share('*' . $_ . '::');
+#S 	    s!::!/!g;
+#S 	    require $_ . '.pm';
+#S 	}
+#S 	$PLP::safe->permit(Opcode::full_opset());
+#S 	$PLP::safe->deny(Opcode::opset(':dangerous'));
+#S     }
+#S     $PLP::safe->reval($code);
+#S }
+#S  my ($r) = @_;
+
+# start()
+#S 	if ($PLP::use_safe) {
+#S 	    PLP::safe_eval($r, $PLP::code);
+#S 	} else {
+#	    eval qq{ package PLP::Script; $PLP::code; };
+#S 	}
+#	PLP::error($@, 1) if $@ and $@ !~ /\cS\cT\cO\cP/;
+#S 	if ($PLP::use_safe) {
+#S 	    PLP::safe_eval($r, '$_->() for reverse @PLP::END');
+#S 	} else {
+#	    eval   { package PLP::Script; $_->() for reverse @PLP::END };
+#S 	}
+#	PLP::error($@, 1) if $@ and $@ !~ /\cS\cT\cO\cP/;
+
+###
